@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2011-2021 Canonical
- * Copyright (C) 2021-2023 Colin Ian King
+ * Copyright (C) 2021-2024 Colin Ian King
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -106,6 +106,7 @@ typedef enum {
 	CPU_FREQ,
 	CPU_FREQ_MIN,
 	CPU_FREQ_MAX,
+	GPU_FREQ,
 	PROC_FORK,
 	PROC_EXEC,
 	PROC_EXIT,
@@ -191,6 +192,7 @@ static const char *cpu_path = "/sys/devices/system/cpu";
 #define OPTS_CPU_FREQ		(0x1000)	/* Average CPU frequency */
 #define OPTS_NO_STATS_HEADINGS	(0x2000)	/* No stats headings */
 #define OPTS_THERMAL_ZONE	(0x4000)	/* Thermal zones */
+#define OPTS_GPU_FREQ		(0x8000)	/* GPU frequency */
 
 #define OPTS_USE_NETLINK	(OPTS_SHOW_PROC_ACTIVITY | \
 				 OPTS_REDO_NETLINK_BUSY |  \
@@ -332,10 +334,10 @@ static const int signals[] = {
 };
 
 /*
- *  strlcpy()
+ *  bsd_strlcpy()
  *	BSD strlcpy
  */
-static size_t strlcpy(char *dst, const char *src, size_t len)
+static size_t bsd_strlcpy(char *dst, const char *src, size_t len)
 {
 	char *d = dst;
 	const char *s = src;
@@ -444,16 +446,20 @@ static const char *cpu_freq_format(const double freq)
 	double scale = 1e18;
 	size_t i;
 
-	for (i = 0; cpu_freq_scale[i].suffix; i++) {
-		if (f < cpu_freq_scale[i].threshold) {
-			suffix = cpu_freq_scale[i].suffix;
-			scale = cpu_freq_scale[i].scale;
-			break;
+	if (freq > 0) {
+		for (i = 0; cpu_freq_scale[i].suffix; i++) {
+			if (f < cpu_freq_scale[i].threshold) {
+				suffix = cpu_freq_scale[i].suffix;
+				scale = cpu_freq_scale[i].scale;
+				break;
+			}
 		}
-	}
 
-	(void)snprintf(buffer, sizeof(buffer), "%5.2f %-3s",
-		f / scale, suffix);
+		(void)snprintf(buffer, sizeof(buffer), "%5.2f %-3s",
+			f / scale, suffix);
+	} else {
+		(void)snprintf(buffer, sizeof(buffer), "  N/A     ");
+	}
 
 	return buffer;
 }
@@ -786,6 +792,19 @@ static void stats_clear_all(stats_t *const stats, const long int n)
 		stats_clear(&stats[i]);
 }
 
+static void stats_gpu_freq_read(stats_t *const stats)
+{
+	uint64_t freq = 0;
+
+	if (file_get_uint64("/sys/class/drm/card0/gt_cur_freq_mhz", &freq) == 0) {
+		stats->value[GPU_FREQ] = (double)freq;
+	} else if (file_get_uint64("/sys/class/graphics/fb0/device/drm/card0/gt_cur_freq_mhz", &freq) == 0) {
+		stats->value[GPU_FREQ] = (double)freq;
+	} else {
+		stats->value[GPU_FREQ] = 0.0;
+	}
+}
+
 static void stats_cpu_freq_read(stats_t *const stats)
 {
 	struct dirent **cpu_list = NULL;
@@ -902,6 +921,9 @@ static int stats_read(stats_t *const stats)
 	if (opts & OPTS_CPU_FREQ)
 		stats_cpu_freq_read(stats);
 
+	if (opts & OPTS_GPU_FREQ)
+		stats_gpu_freq_read(stats);
+
 	return 0;
 }
 
@@ -990,6 +1012,7 @@ static bool stats_gather(
 	res->value[CPU_FREQ] = s2->value[CPU_FREQ];
 	res->value[CPU_FREQ_MIN] = s2->value[CPU_FREQ_MIN];
 	res->value[CPU_FREQ_MAX] = s2->value[CPU_FREQ_MAX];
+	res->value[GPU_FREQ] = s2->value[GPU_FREQ];
 
 	return true;
 }
@@ -1009,9 +1032,9 @@ static void stats_headings(void)
 
 	if (opts & (OPTS_DOMAIN_STATS |
 		    OPTS_THERMAL_ZONE |
-		    OPTS_CPU_FREQ))
+		    OPTS_CPU_FREQ |
+		    OPTS_GPU_FREQ))
 		(void)putchar(' ');
-
 	if (opts & OPTS_DOMAIN_STATS) {
 		for (i = 0; i < power_domains; i++)
 			(void)printf(" %6.6s",
@@ -1023,6 +1046,8 @@ static void stats_headings(void)
 	}
 	if (opts & OPTS_CPU_FREQ)
 		(void)printf(" %9.9s %9.9s %9.9s", "CPU Freq", "Freq Min", "Freq Max");
+	if (opts & OPTS_GPU_FREQ)
+		(void)printf(" %9.9s", "GPU Freq");
 	(void)printf("\n");
 }
 
@@ -1048,6 +1073,8 @@ static void stats_ruler(void)
 			(void)printf(" ------");
 	}
 	if (opts & OPTS_CPU_FREQ)
+		(void)printf(" --------- --------- ---------");
+	if (opts & OPTS_GPU_FREQ)
 		(void)printf(" ---------");
 	(void)printf("\n");
 }
@@ -1134,6 +1161,8 @@ static void stats_print(
 		(void)printf(" %s", cpu_freq_format(s->value[CPU_FREQ_MIN]));
 		(void)printf(" %s", cpu_freq_format(s->value[CPU_FREQ_MAX]));
 	}
+	if (opts & OPTS_GPU_FREQ)
+		(void)printf(" %s", cpu_freq_format(s->value[GPU_FREQ]));
 	(void)printf("\n");
 }
 
@@ -1787,7 +1816,7 @@ static int rapl_get_domains(void)
 					char buf[sizeof(domain_name)];
 
 					(void)snprintf(buf, sizeof(buf), "pkg-%s", domain_name + 8);
-					strlcpy(domain_name, buf, sizeof(domain_name));
+					bsd_strlcpy(domain_name, buf, sizeof(domain_name));
 				}
 
 				if (rapl_domain_unique(domain_name))
@@ -1810,7 +1839,7 @@ static int rapl_get_domains(void)
 	power_domains = MIN(n, MAX_POWER_DOMAINS);
 
 	if (!n)
-		(void)printf("Device does not have any RAPL domains, cannot power measure power usage\n");
+		(void)printf("Device does not have any RAPL domains, cannot measure power usage.\n");
 	return n;
 }
 
@@ -1841,7 +1870,7 @@ static char *power_get_rapl_domain_names(void)
 			break;
 		}
 		names = tmp;
-		(void)strlcpy(names + len, new_name, new_len + 1);
+		(void)bsd_strlcpy(names + len, new_name, new_len + 1);
 		len += new_len;
 	}
 
@@ -2149,7 +2178,7 @@ static cpu_state_t *cpu_state_get(const char *name)
 		free(s);
 		return NULL;
 	}
-	(void)strlcpy(s->name_short, name, len + 1);
+	(void)bsd_strlcpy(s->name_short, name, len + 1);
 	s->name_short[len] = '\0';
 	s->hash_next = cpu_states[h];
 	cpu_states[h] = s;
@@ -2192,10 +2221,10 @@ static cpu_info_t *cpu_info_get(const char *state, const uint32_t cpu_id)
 	(void)memset(buffer, 0, sizeof(buffer));
 	if ((fp = fopen(path, "r")) != NULL) {
 		if (fscanf(fp, "%63s", buffer) != 1)
-			(void)strlcpy(buffer, "unknown", sizeof(buffer));
+			(void)bsd_strlcpy(buffer, "unknown", sizeof(buffer));
 		(void)fclose(fp);
 	} else {
-		(void)strlcpy(buffer, state, sizeof(buffer));
+		(void)bsd_strlcpy(buffer, state, sizeof(buffer));
 	}
 	if ((ci->cpu_state = cpu_state_get(buffer)) == NULL) {
 		free(ci->state);
@@ -2438,7 +2467,7 @@ static int proc_cmdline(
 	}
 
 	if (n < 1) {
-		(void)strlcpy(cmdline, "<unknown>", size);
+		(void)bsd_strlcpy(cmdline, "<unknown>", size);
 		n = 9;
 	}
 
@@ -2913,12 +2942,13 @@ static void show_help(char *const argv[])
 {
 	(void)printf("%s, version %s\n\n", app_name, VERSION);
 	(void)printf("usage: %s [options] [delay [count]]\n", argv[0]);
-	(void)printf("\t-a enable all sampling collection options (-c, -f, -t and -H)\n");
+	(void)printf("\t-a enable all sampling collection options (-c, -f, -g, -t and -H)\n");
 	(void)printf("\t-b redo a sample if a system is busy, considered less than %d%% CPU idle\n", IDLE_THRESHOLD);
 	(void)printf("\t-c show C-State statistics at end of the run\n");
 	(void)printf("\t-d specify delay before starting, default is %" PRId32 " seconds\n", start_delay);
 	(void)printf("\t-D show RAPL domain power measurements (enables -R option)\n");
 	(void)printf("\t-f show average CPU frequency\n");
+	(void)printf("\t-g show average GPU frequency\n");
 	(void)printf("\t-h show help\n");
 	(void)printf("\t-H show spread of measurements with power histogram\n");
 	(void)printf("\t-i specify CPU idle threshold, used in conjunction with -b\n");
@@ -2949,9 +2979,9 @@ int main(int argc, char * const argv[])
 
 	for (;;) {
 #if defined(POWERSTAT_X86)
-		int c = getopt(argc, argv, "abd:cDfhHi:nprszStR");
+		int c = getopt(argc, argv, "abd:cDfghHi:nprszStR");
 #else
-		int c = getopt(argc, argv, "abd:cDfhHi:nprszSt");
+		int c = getopt(argc, argv, "abd:cDfghHi:nprszSt");
 #endif
 		if (c == -1)
 			break;
@@ -2959,6 +2989,7 @@ int main(int argc, char * const argv[])
 		case 'a':
 			opts |= (OPTS_CSTATES |
 				 OPTS_CPU_FREQ |
+				 OPTS_GPU_FREQ |
 				 OPTS_HISTOGRAM |
 				 OPTS_THERMAL_ZONE);
 			break;
@@ -2987,6 +3018,9 @@ int main(int argc, char * const argv[])
 		case 'f':
 			opts |= OPTS_CPU_FREQ;
 			break;
+		case 'g':
+			opts |= OPTS_GPU_FREQ;
+			break;
 		case 'h':
 			show_help(argv);
 			exit(EXIT_SUCCESS);
